

open Smt_sorts;;
open Smt_tables;;
open Smt_helper_theory;;
open Smt_parser_theory;;
open Smt_lexer_theory;;
open Smt_helper_logic;;
open Smt_parser_logic;;
open Smt_lexer_logic;;
open Smt_helper;;
open Smt_parser;;
open Smt_lexer;;
open Smt_util;;

module StringMap = Map.Make (String);;

type key_environment =
    { variables : string -> string;
      fvariables : string -> string;
      formulas : formula StringMap.t;
      terms : term StringMap.t;
      functions : string -> string;
      polarity : int;
      underneathQuant : bool};;


let init_environment =
  let id x = x in
  let func x = x in
  { variables = id;
    fvariables = id;
    formulas = StringMap.empty;
    terms = StringMap.empty;
    functions = func;
    polarity = -1;
    underneathQuant = false }
;;



let print_predFunc o s =
  match s with
    "~" -> Printf.fprintf o "neg"
  | "-" -> Printf.fprintf o "sub"
  | "+" -> Printf.fprintf o "add"
  | "*" -> Printf.fprintf o "mul"
  | "<=" -> Printf.fprintf o "leq"
  | "<" -> Printf.fprintf o "lt"
  | ">=" -> Printf.fprintf o "geq"
  | ">" -> Printf.fprintf o "gt"
  | "null" -> Printf.fprintf o "nullSMT"
  | anything -> Printf.fprintf o "%s" anything
;;

let print_predFuncApp o s terms =
  match s with
    "~" -> Printf.fprintf o "neg"
  | "-" -> Printf.fprintf o "sub"
  | "+" -> Printf.fprintf o "add"
  | "*" -> Printf.fprintf o "mul"
  | "<=" -> Printf.fprintf o "leq"
  | "<" -> Printf.fprintf o "lt"
  | ">=" -> Printf.fprintf o "geq"
  | ">" -> Printf.fprintf o "gt"
  | "null" -> Printf.fprintf o "nullSMT"
  | anything -> Printf.fprintf o "%s" anything
;;





let declared_symbols =
  object
    val mutable symbol_counter = 0

    val mutable variables : string list = []
    val mutable fvariables : string list = []

    method new_sym =
      let sym = "v_princess_" ^ (string_of_int symbol_counter) in
      symbol_counter <- (symbol_counter + 1);
      sym

    method add_var sym = variables <- (sym::variables)
    method add_fvar sym = fvariables <- (sym::fvariables)

    method print_declarations o =
      Printf.fprintf o "\\functions {\n";
      let print x =
	Printf.fprintf o "int %s;\n" x in
      List.iter print variables;
      Printf.fprintf o "}\n";
      Printf.fprintf o "\\predicates {\n";
      let print x =
	print_predFunc o x;
	Printf.fprintf o ";\n" in
      List.iter print fvariables;
      Printf.fprintf o "}\n";
  end
;;


let new_variable env name =
  let new_v = declared_symbols#new_sym in
  let new_variables x = if x = name then new_v else env.variables x in
  { env with variables = new_variables }
;;

let new_fvariable env name =
  let new_v = declared_symbols#new_sym in
  let new_variables x = if x = name then new_v else env.fvariables x in
  { env with fvariables = new_variables }
;;

let neg_polarity env =
  { env with polarity = -env.polarity }
;;

let zero_polarity env =
  { env with polarity = 0 }
;;

let add_formula env name form =
  { env with formulas = StringMap.add name form env.formulas }
;;

let add_term env name term =
  { env with terms = StringMap.add name term env.terms }
;;

let pass_quantifier env =
  { env with underneathQuant = true }
;;


let remove_last list =
  List.rev (List.tl (List.rev list))
;;

let last list =
  List.hd (List.rev list)
;;


let print_sort o s =
  match s with
    "Int" -> Printf.fprintf o "int"
  | anything -> Printf.fprintf o "%s" anything
;;

let argument_sorts_decl o sorts =
  if List.length sorts > 0 then begin
    Printf.fprintf o "(";
    let with_comma t = print_sort o t; Printf.fprintf o ", " in
    List.iter with_comma (remove_last sorts);
    print_sort o (last sorts);
    Printf.fprintf o ")"
  end
;;




let bind_variable o env binder =
  match binder with (var_name, var_sort) ->
    let new_env = new_variable env var_name in
    let new_v = new_env.variables var_name in
    print_sort o var_sort;
    Printf.fprintf o " %s" new_v;
    new_env
;;

let bind_variable_semi o env binder =
  let new_env = bind_variable o env binder in
  Printf.fprintf o "; ";
  new_env  
;;


let bind_variables o env binders =
  match binders with
    [var] ->
      bind_variable_semi o env var
  | (var::var2::_) ->
      Printf.fprintf o "(";
      let first_binders = remove_last binders in
      let last_binder = last binders in
      let new_env1 = List.fold_left (bind_variable_semi o) env first_binders in
      let new_env = bind_variable o new_env1 last_binder in
      Printf.fprintf o ")\n";
      new_env
  | anything ->
      Printf.fprintf o "Expected a variable\n";
      env
;;


let add_ite_definitions itedefs form polarity =
  let add_def d f =
    match d with
      (sym, guard, term1, term2) ->
	let eq1 = EQUALS([T(VAR(sym)); term1]) in
	let eq2 = EQUALS([T(VAR(sym)); term2]) in
	let ite = IF_THEN_ELSE(guard, F(eq1), F(eq2)) in
	if polarity > 0 then
	  QUANT(FORALL, [(sym, "Int")], F(IMPLIES(F(ite), F(f))))
	else
	  QUANT(EXISTS, [(sym, "Int")], F(AND([F(ite); F(f)]))) in
  (List.fold_right add_def itedefs form)
;;


let rec expand_ite_formula2 form polarity =
  if polarity == 0 then
    failwith "polarity zero is not supported at this point";    
  match form with
    EQUALS(terms) ->
      let subres = (List.map expand_ite_term terms) in
      let (newterms, itedefs) = (List.split subres) in
      (add_ite_definitions (List.flatten itedefs) (EQUALS(newterms)) polarity)
  | DISTINCT(terms) ->
      let subres = (List.map expand_ite_term terms) in
      let (newterms, itedefs) = (List.split subres) in
      (add_ite_definitions (List.flatten itedefs) (DISTINCT(newterms)) polarity)
  | PRED(sym, terms) ->
      let subres = (List.map expand_ite_term terms) in
      let (newterms, itedefs) = (List.split subres) in
      (add_ite_definitions (List.flatten itedefs) (PRED(sym, newterms)) polarity)
  | _ -> (failwith "unexpected case")
	

and expand_ite_term term =
  match term with
    T(t) ->
      let (newterm, itedefs) = expand_ite_term2 t in
      (T(newterm), itedefs)
  | TA(t,_) ->
      let (newterm, itedefs) = expand_ite_term2 t in
      (T(newterm), itedefs)

and expand_ite_term2 term =
  match term with
    VAR(sym) -> (VAR(sym), [])
  | FUN(sym, terms) ->
      let subres = (List.map expand_ite_term terms) in
      let (newterms, itedefs) = (List.split subres) in
      (FUN(sym, newterms), (List.flatten itedefs))
  | NUMERAL(i) -> (NUMERAL(i), [])
  | ITE(form, term1, term2) ->
      let var = declared_symbols#new_sym in
      let (newterm1, itedefs1) = expand_ite_term term1 in
      let (newterm2, itedefs2) = expand_ite_term term2 in
      (VAR(var), (List.flatten [itedefs1; itedefs2; [(var, form, newterm1, newterm2)]]))
;;


let rec simple_formula form =
  match form with
    F(t) -> simple_formula2 t
  | FA(t, _) -> simple_formula2 t

and simple_formula2 form =
  match form with
    NOT(f) -> simple_formula f
  | AND([f]) -> simple_formula f
  | OR([f]) -> simple_formula f
  | FVAR(sym) -> true
  | PRED(rawsym, rawterms) -> true
  | EQUALS(terms) -> true
  | DISTINCT(terms) -> (List.length terms) <= 2
  | FALSE_PRED -> true
  | TRUE_PRED -> true
  | _ -> false
;;

let rec simple_term term =
  match term with
    T(t) -> simple_term2 t
  | TA(t,_) -> simple_term2 t

and simple_term2 term =
  match term with
    VAR(_) -> true
  | NUMERAL(_) -> true
  | _ -> false
;;


let rec formula2_key o env form =
  match form with
    FLET(sym, form1, form2) ->
      if env.underneathQuant then
	failwith "flet underneath quantifiers is not supported";	
      Printf.fprintf o "(";
      if simple_formula form1 then
	let new_env = add_formula env sym form1 in
	formula_key o new_env form2
      else begin
	let new_env = new_fvariable env sym in
	declared_symbols#add_fvar (new_env.fvariables sym);
	if env.polarity > 0 then
	  formula2_key o new_env (IMPLIES(F(IFF [form1; F(FVAR(sym))]), form2))
	else if env.polarity < 0 then
	  formula2_key o new_env (AND([F(IFF [form1; F(FVAR(sym))]); form2]))
	else
	  failwith "polarity zero is not supported at this point"
      end;
      Printf.fprintf o ")"
  | LET(sym, term, subform) ->
      Printf.fprintf o "(";
      if simple_term term then
	let new_env = add_term env sym term in
	formula_key o new_env subform
      else begin
	let eq = F(EQUALS([T(VAR(sym)); term])) in
	if env.polarity > 0 then
	  formula2_key o env (QUANT(FORALL, [(sym, "Int")], F(IMPLIES(eq, subform))))
	else if env.polarity < 0 then
	  formula2_key o env (QUANT(EXISTS, [(sym, "Int")], F(AND([eq; subform]))))
	else
	  failwith "polarity zero is not supported at this point"
      end;
      Printf.fprintf o ")"
  | QUANT(q, vars, subform) ->
      Printf.fprintf o "(";
      if env.underneathQuant == false &&
	 (q == FORALL && env.polarity > 0 ||
	 q == EXISTS && env.polarity < 0) then
	(* skolemise *)
	let declare_var env x =
	  match x with
	    (var_name, var_sort) ->
	      let new_env = new_variable env var_name in
	      declared_symbols#add_var (new_env.variables var_name);
	      new_env
	in
	
	let new_env = (List.fold_left declare_var env vars) in
	formula_key o new_env subform	

      else begin

	begin
          match q with  FORALL -> Printf.fprintf o "\\forall "
                      | EXISTS -> Printf.fprintf o "\\exists "
	end;
	let new_env2 = bind_variables o env vars in
	let new_env = pass_quantifier new_env2 in
	formula_key o new_env subform;
      end;
      Printf.fprintf o ")"
  | EQUALS(rawterms) -> begin
      match expand_ite_formula2 (EQUALS(rawterms)) env.polarity with
	EQUALS(terms) ->
	  if List.length terms < 2 then
	    Printf.fprintf o "true"
	  else begin
	    let lhs = remove_last terms in
	    let rhs = List.tl terms in
	    let print_eq a b =
	      Printf.fprintf o "(";
	      term_key o env a;
	      Printf.fprintf o ") = (";
	      term_key o env b;
	      Printf.fprintf o ")" in
	    let print_eq_and a b =
	      print_eq a b;
	      Printf.fprintf o " & " in
	    List.iter2 print_eq_and (List.tl lhs) (List.tl rhs);
	    print_eq (List.hd lhs) (List.hd rhs)
	  end
      | x -> formula2_key o env x
  end
  | NOT(subform) ->
      Printf.fprintf o "!(";
      formula_key o (neg_polarity env) subform;
      Printf.fprintf o ")"      
  | AND(forms) -> fold_formulas o env "true" " & " forms
  | OR(forms) -> fold_formulas o env "false" " | " forms
  | IFF(forms) ->
      if (List.length forms) == 2 then
	let for1 = (List.hd forms) in
	let for2 = (List.nth forms 1) in
	formula2_key o env (IF_THEN_ELSE(for1, for2, F(NOT(for2))))
      else
	fold_formulas o (zero_polarity env) "true" " <-> " forms
  | XOR(forms) ->
      Printf.fprintf o "(";
      fold_formulas o (zero_polarity env) "false" ") <-> !(" forms;
      Printf.fprintf o ")"
  | IMPLIES(form1,form2) ->
      Printf.fprintf o "((";
      formula_key o (neg_polarity env) form1;
      Printf.fprintf o ") -> (";
      formula_key o env form2;
      Printf.fprintf o "))"
  | FVAR(sym) ->
      if StringMap.mem sym env.formulas then begin
	Printf.fprintf o "(";
	formula_key o env (StringMap.find sym env.formulas);
	Printf.fprintf o ")"
      end else begin
	Printf.fprintf o "%s()" (env.fvariables sym)
      end
  | PRED(rawsym, rawterms) -> begin
      match expand_ite_formula2 (PRED(rawsym, rawterms)) env.polarity with
	PRED(sym, terms) ->
	  infix_funPredApp o env sym terms true
      | x -> formula2_key o env x
  end
  | IF_THEN_ELSE(form, form1, form2) ->
      let imp1 = IMPLIES(form, form1) in
      let imp2 = IMPLIES(F(NOT(form)), form2) in
      let a = AND([F(imp1); F(imp2)]) in
      formula2_key o env a
  | FALSE_PRED -> Printf.fprintf o "false"
  | TRUE_PRED -> Printf.fprintf o "true"
  | DISTINCT(rawterms) -> begin
      match expand_ite_formula2 (DISTINCT(rawterms)) env.polarity with
	DISTINCT(terms) ->
	  Printf.fprintf o "(";
	  print_unequal_list o env terms;
	  Printf.fprintf o ")"
      | x -> formula2_key o env x
  end

and formula_key o env form =
  match form with
    F(f) -> formula2_key o env f
  | FA(f,_) -> formula2_key o env f

and term_key o env term =
  match term with
    T(t) -> term2_key o env t
  | TA(t,_) -> term2_key o env t

and term2_key o env term =
  match term with
    VAR(sym) ->
      if StringMap.mem sym env.terms then begin
	Printf.fprintf o "(";
	term_key o env (StringMap.find sym env.terms);
	Printf.fprintf o ")"
      end else begin
	Printf.fprintf o "%s" (env.variables sym)
      end
  | FUN(sym, terms) ->
      infix_funPredApp o env sym terms false
  | NUMERAL(i) -> Printf.fprintf o "%s" (Big_int.string_of_big_int i)
  | ITE(form, term1, term2) ->
      Printf.fprintf o "\\if (";
      formula_key o env form;
      Printf.fprintf o ") \\then (";
      term_key o env term1;
      Printf.fprintf o ") \\else (";
      term_key o env term2;
      Printf.fprintf o ")"

and fold_formulas o env def conn forms =
  Printf.fprintf o "(";
  if List.length forms < 1 then begin
    Printf.fprintf o "%s" def
  end else begin
    let print_one a =
      formula_key o env a;
      Printf.fprintf o "%s" conn in
    List.iter print_one (List.tl forms);
    formula_key o env (List.hd forms)
  end;
  Printf.fprintf o ")"

and fold_terms o env def conn terms =
  Printf.fprintf o "(";
  if List.length terms < 1 then begin
    Printf.fprintf o "%s" def
  end else begin
    let print_one a =
      term_key o env a;
      Printf.fprintf o "%s" conn in
    List.iter print_one (List.tl terms);
    term_key o env (List.hd terms)
  end;
  Printf.fprintf o ")"

and arguments_key o env terms alwaysParen =
  if List.length terms > 0 then begin
    Printf.fprintf o "(";
    let with_comma t = term_key o env t; Printf.fprintf o ", " in
    let rev_terms = List.rev terms in
    List.iter with_comma (remove_last terms);
    term_key o env (List.hd rev_terms);
    Printf.fprintf o ")"
  end else if alwaysParen then begin    
    Printf.fprintf o "()";
  end

and infix_funPredApp o env s terms alwaysParen =
  match s with
    "-" | "<=" | "<" | ">=" | ">" ->
           print_bin_infix o env s (List.hd terms) (List.nth terms 1)
  | "+" -> fold_terms o env "0" "+" terms
  | "*" -> fold_terms o env "1" "*" terms
  | "~" -> print_bin_infix o env "-" (T(NUMERAL(Big_int.zero_big_int))) (List.hd terms)
  | anything ->
           Printf.fprintf o "%s" anything;
           arguments_key o env terms alwaysParen

and print_bin_infix o env s t1 t2 =
  Printf.fprintf o "((";
  term_key o env t1;
  Printf.fprintf o ") %s (" s;
  term_key o env t2;
  Printf.fprintf o "))"  

and print_unequal_step o env t terms =
  let one_unequal t2 =
    Printf.fprintf o " & ";
    print_bin_infix o env "!=" t t2 in
  List.iter one_unequal terms

and print_unequal_list o env terms =
  match terms with
    [] -> Printf.fprintf o "true"
  | [t] -> Printf.fprintf o "true"
  | (t::rem) ->
      print_unequal_list o env rem;
      print_unequal_step o env t rem
;;


let is_extrasorts a =
  match a with
    EXTRASORTS b -> true
  | _ -> false;;

let get_extrasorts attribute =
  match attribute with
    EXTRASORTS(s) -> s
  | _ -> [];;

let get_extrasorts_from_benchmark benchmark =
  List.flatten (List.map get_extrasorts (List.filter is_extrasorts
					   (get_benchmark_attributes benchmark)));;

let is_extrafuns a =
  match a with
    EXTRAFUNS b -> true
  | _ -> false;;

let get_extrafuns attribute =
  match attribute with
    EXTRAFUNS(s) -> s
  | _ -> [];;

let get_extrafuns_from_benchmark benchmark =
  List.flatten (List.map get_extrafuns (List.filter is_extrafuns
					   (get_benchmark_attributes benchmark)));;

let is_extrapreds a =
  match a with
    EXTRAPREDS b -> true
  | _ -> false;;

let get_extrapreds attribute =
  match attribute with
    EXTRAPREDS(s) -> s
  | _ -> [];;

let get_extrapreds_from_benchmark benchmark =
  List.flatten (List.map get_extrapreds (List.filter is_extrapreds
					   (get_benchmark_attributes benchmark)));;


let print_problem o assumptions formulas =
  Printf.fprintf o "\\problem {\n";
  fold_formulas o init_environment "true" " & " assumptions;
  Printf.fprintf o "\n &\n";
  fold_formulas o init_environment "true" " & " formulas;
  Printf.fprintf o "\n -> false\n";
  Printf.fprintf o "}\n"
;;


let print_sorts o sorts =
  Printf.fprintf o "\\sorts {\n";

  Printf.fprintf o "Array;\n";

  List.iter (Printf.fprintf o "%s;\n") sorts;
  Printf.fprintf o "}\n"
;;


let print_funs o funs =
  Printf.fprintf o "\\functions {\n";

(*  Printf.fprintf o "int select(Array, int);\n";
  Printf.fprintf o "Array store(Array, int, int);\n"; *)

  let print x =
    match x with FUNC_SIG(name, args, result, _) ->
      print_sort o result;
      Printf.fprintf o " ";
      print_predFunc o name;
      argument_sorts_decl o args;
      Printf.fprintf o ";\n" in
  List.iter print funs;
  Printf.fprintf o "}\n"
;;

let print_preds o preds =
  Printf.fprintf o "\\predicates {\n";
  let print x =
    match x with PRED_SIG(name, args, _) ->
      print_predFunc o name;
      argument_sorts_decl o args;
      Printf.fprintf o ";\n" in
  List.iter print preds;
  Printf.fprintf o "}\n"
;;


let print_Array_axioms o =
  Printf.fprintf o "\\schemaVariables {\n";
  Printf.fprintf o "  \\term Array ar;\n";
  Printf.fprintf o "  \\term int element, index0, index1;\n";
  Printf.fprintf o "}\n";
  Printf.fprintf o "\\rules {\n";
  Printf.fprintf o "  arrayAxiom {\n";
  Printf.fprintf o "    \\find (select(store(ar, index0, element), index1))\n";
  Printf.fprintf o "    \\replacewith (\\if (index0 = index1)\n";
  Printf.fprintf o "                  \\then (element)\n";
  Printf.fprintf o "                  \\else (select(ar, index1)))\n";
  Printf.fprintf o "    \\heuristics (simplify)\n";
  Printf.fprintf o "  };\n";
  Printf.fprintf o "}\n";
;;


let convert o filename = 
  let benchmark = Smt_util.parse_benchmark_file filename in
  let formulas = Smt_helper.get_formula_from_benchmark benchmark in
  let assumptions = Smt_helper.get_assumption_from_benchmark benchmark in
  let extrasorts = get_extrasorts_from_benchmark benchmark in
  let extrafuns = get_extrafuns_from_benchmark benchmark in
  let extrapreds = get_extrapreds_from_benchmark benchmark in

(*  print_sorts o extrasorts; *)
  print_funs o extrafuns;
  print_preds o extrapreds;
(*  print_Array_axioms o; *)
  print_problem o assumptions formulas;
  declared_symbols#print_declarations o
;;
  


if Array.length Sys.argv = 3 then begin
  let o = open_out Sys.argv.(2) in
  convert o Sys.argv.(1);
  close_out o
end else begin
  Printf.printf "Usage: convert_key <inputfile> <outputfile>\n";
end
